from typing import Dict, Union
import os


class ServerConfig:
    __slots__ = ["backend", "ip", "port"]

    def __init__(self, backend: str, ip: str, port: Union[str, int]):
        self.backend = backend
        self.ip = ip
        self.port = port


def parse_server_config(config: Union[Dict, str]) -> ServerConfig:
    if isinstance(config, dict):
        return ServerConfig(config["backend"], config["ip"], config["port"])
    if not (isinstance(config, str) and os.path.exists(config)):
        raise TypeError(f"SERVER_CONFIG must be dict or path string - {config}")

    d = {}
    with open(config) as f:
        for line in f.readlines():
            if not line:
                continue

            key, value = [i.strip() for i in line.split("=", 1)]
            d[key.lower()] = value

    return ServerConfig(d["backend"], d["ip"], d["port"])



from typing import Any
from abc import ABC, abstractmethod

from lodis_py.client import AsyncLodisClient


class Wrapper(ABC):
    @abstractmethod
    async def exists(self, key) -> bool:
        """Check whether the key exists in the queue"""
        raise NotImplementedError

    @abstractmethod
    async def push_left(self, key, value) -> None:
        """Add tasks to the left of the queue"""
        raise NotImplementedError

    @abstractmethod
    async def pushnx_left(self, key, value) -> None:
        """Add tasks to the left of the queue only if item does not exists in the queue"""
        raise NotImplementedError

    @abstractmethod
    async def push(self, key, value) -> None:
        """Add tasks to the right of the queue"""
        raise NotImplementedError

    @abstractmethod
    async def pushnx(self, key, value) -> None:
        """Add tasks to the right of the queue only if item does not exists in the queue"""
        raise NotImplementedError

    @abstractmethod
    async def pop_left(self) -> Any:
        """Take out an task from the left of the queue"""
        raise NotImplementedError

    @abstractmethod
    async def pop(self) -> Any:
        """Take out an task from the right of the queue"""
        raise NotImplementedError

    @abstractmethod
    async def size(self) -> int:
        """The size of the queue"""
        raise NotImplementedError

    @abstractmethod
    async def finish(self, key) -> None:
        """Set the task which has the key to finished"""
        raise NotImplementedError

    @abstractmethod
    async def finished(self, key) -> bool:
        """Check whether a task is finished"""
        raise NotImplementedError

    @abstractmethod
    async def unfinish(self, key) -> None:
        """Set the task which has the key to unfinished"""
        raise NotImplementedError

    @abstractmethod
    async def clear_unfinished_tasks(self) -> None:
        """Clear all unfinished tasks from the queue"""
        raise NotImplementedError

    @abstractmethod
    async def clear_finished_tasks(self) -> None:
        """Clear all finished tasks"""
        raise NotImplementedError

    async def clear_all(self) -> None:
        """Clear all unfinished and finished tasks"""

        await self.clear_unfinished_tasks()
        await self.clear_finished_tasks()


class LodisWrapper(Wrapper):
    def __init__(self, namespace, ip, port, loop=None):
        self._lodis = AsyncLodisClient(namespace, ip, port, loop=loop)

    # ArrayMap
    async def exists(self, key) -> bool:
        resp = await self._lodis.aexists(key)
        return resp.value()  # type: ignore

    async def push_left(self, *pairs) -> None:
        await self._lodis.alpush(*pairs)

    async def pushnx_left(self, *pairs) -> None:
        await self._lodis.alpushnx(*pairs)

    async def push(self, *pairs) -> None:
        await self._lodis.arpush(*pairs)

    async def pushnx(self, *pairs) -> None:
        await self._lodis.arpushnx(*pairs)

    async def pop_left(self) -> Any:
        resp = await self._lodis.alpop()
        return resp.value()  # type: ignore

    async def pop(self) -> Any:
        resp = await self._lodis.arpop()
        return resp.value()  # type: ignore

    async def size(self) -> int:
        resp = await self._lodis.alen()
        return resp.value()  # type: ignore

    # Map
    async def finish(self, *keys) -> None:
        pairs = [(key, b"") for key in keys]
        await self._lodis.hmset(*pairs)

    async def finished(self, key) -> bool:
        resp = await self._lodis.hexists(key)
        return resp.value()  # type: ignore

    async def unfinish(self, key) -> None:
        await self._lodis.hdel(key)

    async def clear_unfinished_tasks(self) -> None:
        await self._lodis.arm()

    async def clear_finished_tasks(self) -> None:
        await self._lodis.hrm()



from typing import Any
import asyncio
import logging
from functools import wraps

from .adapter import LodisWrapper, Wrapper
from .config import ServerConfig
from .error import UnknownServer

logger = logging.getLogger(__name__)


def handle_exception(gen):
    @wraps(gen)
    async def wrapper(*args, **kwargs):
        while True:
            try:
                # TODO: Handle timeout at here
                return await gen(*args, **kwargs)
            except Exception as err:
                logger.error("Server error: %s, future: %s", err, gen)
                await asyncio.sleep(2)

    return wrapper


class Server(Wrapper):
    def __init__(
        self,
        namespace: str,
        config: ServerConfig,
        loop: asyncio.AbstractEventLoop = None,
    ):
        self._namespace = namespace
        self._config = config
        self._loop = loop
        self._server = self._select_server(namespace, config, loop)

    def _select_server(
        self,
        namespace: str,
        config: ServerConfig,
        loop: asyncio.AbstractEventLoop = None,
    ):
        backend = config.backend
        if backend == "lodis":
            return LodisWrapper(namespace, config.ip, config.port, loop=loop)
        else:
            raise UnknownServer(backend)

    @handle_exception
    async def exists(self, key) -> bool:
        return await self._server.exists(key)

    @handle_exception
    async def push_left(self, *pairs) -> None:
        await self._server.push_left(*pairs)

    @handle_exception
    async def pushnx_left(self, *pairs) -> None:
        await self._server.pushnx_left(*pairs)

    @handle_exception
    async def push(self, *pairs) -> None:
        await self._server.push(*pairs)

    @handle_exception
    async def pushnx(self, *pairs) -> None:
        await self._server.pushnx(*pairs)

    @handle_exception
    async def pop_left(self) -> Any:
        return await self._server.pop_left()

    @handle_exception
    async def pop(self) -> Any:
        return await self._server.pop()

    @handle_exception
    async def size(self) -> int:
        return await self._server.size()

    @handle_exception
    async def finish(self, *keys) -> None:
        await self._server.finish(*keys)

    @handle_exception
    async def finished(self, key) -> bool:
        return await self._server.finished(key)

    @handle_exception
    async def unfinish(self, key) -> None:
        await self._server.unfinish(key)

    @handle_exception
    async def clear_unfinished_tasks(self) -> None:
        await self._server.clear_unfinished_tasks()

    @handle_exception
    async def clear_finished_tasks(self) -> None:
        await self._server.clear_finished_tasks()



import logging
import asyncio

from .loop import ChihuoLoop
from .signal_handlers import set_signal_handlers

logger = logging.getLogger(__name__)


async def _event_alive():
    """Keep event loop alive

    Event loop could be dead if all tasks are keeping.
    The task keeps event loop to listen for events. Because it be woke up periodically.
    """

    while True:
        await asyncio.sleep(1)


def _instantiate_factories(*classes_or_instances):
    """Instantiate factories"""

    factories = []
    for obj in classes_or_instances:
        if isinstance(obj, ChihuoLoop):
            if obj.NAME is None:
                raise TypeError("factory.NAME must be given")
            factories.append(obj)
        else:
            if type(obj) is not type:
                raise TypeError("factory must be a class or instance of ChihuoLoop")
            if not issubclass(obj, ChihuoLoop):
                raise TypeError(
                    "factory must be a subclass of ChihuoLoop or instance of ChihuoLoop"
                )
            if obj.NAME is None:
                raise TypeError("factory.NAME must be given")

            factories.append(obj())
    return factories


def run(*classes_or_instances):
    """Instantiate factories and run them"""

    if not classes_or_instances:
        logger.error("No provide factory class or instance")

    assert classes_or_instances, "No provide factory class or instance"

    loop = asyncio.get_event_loop()

    # Keep event loop alive
    loop.create_task(_event_alive())

    factories = _instantiate_factories(*classes_or_instances)
    logger.info("Find factories: %s", [factory.NAME for factory in factories])

    set_signal_handlers(factories, loop)

    # Run factories
    for factory in factories:
        factory._run_()

    loop.run_forever()



from typing import Union
from enum import Enum

SERVER_DEFAULT_CONFIG_PATH = "server.config"

TaskId = Union[str, bytes]


class TaskType(Enum):
    Task = "task"
    Start = "start"
    TaskLoop = "task_loop"
    SendBack = "send_back"
    Handler = "handler"
    Final = "final"


class Direction(Enum):
    Forward = 0
    Reverse = 1



from typing import Any
import types
import json
import asyncio


def make_ctrl_queue(concurrency: int, loop: asyncio.AbstractEventLoop = None):
    queue: asyncio.Queue = asyncio.Queue(concurrency, loop=loop)
    for _ in range(concurrency):
        queue._put(1)
    return queue


def serialize_json(task: Any):
    return json.dumps(task, separators=(",", ":"), ensure_ascii=False)


async def aretry(async_func, times: int, *args, **kwargs):
    """Retry an asynchronous function for `times` times

    Params:
    async_func: an asynchronous function
    args: arguments for async_func
    kwargs: keyword arguments for async_func
    times: retry times
    """

    assert isinstance(async_func, (types.FunctionType, types.MethodType))

    # If times <= 0, we retry the async_func until it returns
    if times < 1:
        times = 1 << 31

    for i in range(times):
        try:
            return await async_func(*args, **kwargs)
        except Exception as err:
            if i + 1 == times:
                raise err



from typing import Iterable, Optional, Tuple, Dict, Any
import logging
import asyncio
import functools
import signal
import copy

from .common import TaskType, TaskId, Direction
from .loop import ChihuoLoop

logger = logging.getLogger(__name__)

IS_ON_TERM = False
IS_ON_INT = False


async def cancel_all_running_tasks() -> None:
    """Cancel all running tasks, then send these tasks back to backend"""

    while True:
        # Cancel all running async tasks in the loop
        tasks = [t for t in asyncio.all_tasks() if t.get_name() != TaskType.Final.value]

        logger.info("all tasks number: %s", len(tasks))

        task_loop = [t for t in tasks if t.get_name() == TaskType.TaskLoop.value]
        needed_cancel = [t for t in tasks if t.get_name() != TaskType.TaskLoop.value]

        for task in needed_cancel:
            task.cancel()

        if len(needed_cancel) == 0:
            # Finally, cancel the `ChihuoLoop._task_loop`
            for task in task_loop:
                task.cancel()
            break

        await asyncio.sleep(1)


async def send_back_running_tasks(
    factory_and_tasks: Iterable[Tuple[ChihuoLoop, Dict[TaskId, Any]]]
) -> None:
    logger.info("[send back task]: START")

    for factory, tasks in factory_and_tasks:
        for task_id, task in tasks.items():
            logger.info(
                "[Send back task]: factory: %s, task: %s",
                factory.__class__.__name__,
                (task_id, task),
            )
            await factory.add_task((task_id, task), direction=Direction.Reverse)

    logger.info("[send back task]: END")


async def teardown(
    factories: Iterable[ChihuoLoop], loop: asyncio.AbstractEventLoop
) -> None:
    logger.info("[teardown]: START")

    # Send back running tasks
    running_tasks = [
        (factory, copy.deepcopy(factory._running_tasks)) for factory in factories
    ]
    await send_back_running_tasks(running_tasks)

    # Spawn the final task
    await cancel_all_running_tasks()

    loop.stop()

    logger.info("[teardown]: END")


def handle_stop(
    signum,
    frame,
    factories: Iterable[ChihuoLoop] = [],
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> None:
    """The handler of INT and TERM signals

    First, we set all factories to be stoped at next event loop.
    Then, we cancel all running tasks which will be sent back to backend.
    """

    assert loop, "The Event Loop can not be None"

    logger.info("[handle_stop]: signal: %s", signum)

    global IS_ON_INT
    if IS_ON_TERM or IS_ON_INT:
        return

    # Stop the factory task loop
    for factory in factories:
        factory.stop = True

    loop.create_task(teardown(factories, loop), name=TaskType.Final.value)

    IS_ON_INT = True


def set_signal_handlers(
    factories: Iterable[ChihuoLoop], loop: asyncio.AbstractEventLoop
) -> None:
    """Set signal handlers for `factories`"""

    signal.signal(
        signal.SIGTERM,
        functools.partial(handle_stop, factories=factories, loop=loop),
    )
    signal.signal(
        signal.SIGINT,
        functools.partial(handle_stop, factories=factories, loop=loop),
    )



from typing import IO, Union
import os
import logging

DEFAULT_FMT = "%(asctime)-15s | %(levelname)s | %(module)s: %(message)s"
DEFAULT_FILENAME = "log/log.log"


def get_logger(
    name: str = None,
    fmt: str = DEFAULT_FMT,
    filename: str = DEFAULT_FILENAME,
    stream: IO[str] = None,
    level: int = logging.INFO,
):
    logger = logging.getLogger(name)
    logger.setLevel(level)

    handler: Union[logging.StreamHandler, logging.FileHandler]

    if filename:
        dirt = os.path.dirname(filename)
        if dirt and not os.path.exists(dirt):
            os.makedirs(dirt)

        handler = logging.FileHandler(filename)
    else:
        handler = logging.StreamHandler(stream=stream)

    handler.setFormatter(logging.Formatter(fmt))
    logger.addHandler(handler)

    return logger



from typing import Union, Dict, Tuple, Optional, Any, cast
import traceback
import logging
import asyncio
import json as stdjson
import types

from .server import Server
from .common import SERVER_DEFAULT_CONFIG_PATH, TaskType, TaskId, Direction
from .config import parse_server_config
from .util import serialize_json

logger = logging.getLogger(__name__)


class ChihuoLoop:
    """Chihuo async loop abstract class

    The class handles all asynchronous tasks which are created
    by the `make_task` method of its subclass.

    The class connects to a Chihuo server to persist all tasks.
    All running tasks are received from the backend server.
    So, for adding a task to the loop, subclass must use the
    `add_task` method to add the task to the backend server,
    then the loop can automaticly get a task from the backend
    server. The received task is passed to `make_task` to make
    an asynchronous generator which is added to the asynchronous
    loop.

    The backend server can be regarded as a bidirectional queue.

    Params:

    NAME: The unique project name (MUST BE ascii characters)
        It is the key to define the task queue of the project in
        backend server.

    CONCURRENCY: The number of all concurrent tasks

    SERVER_CONFIG: The configuration of the backend server
        It is a dict or a path to the configuration file.
        Default is the path 'chdb.config'
    """

    NAME: Optional[str] = None
    CONCURRENCY: int = 1
    SERVER_CONFIG: Union[str, Any] = SERVER_DEFAULT_CONFIG_PATH

    def __init__(self, run_forever: bool = True):
        if not self.NAME:
            logger.warning("NAME is missing")

        assert self.NAME, "NAME is missing"
        assert self.CONCURRENCY > 0, "CONCURRENCY must be > 0"

        logger.info(
            """
Initiate %s loop:
    NAME: %s
    CONCURRENCY: %s
    SERVER_CONFIG: %s""",
            self._cls_name,
            self.NAME,
            self.CONCURRENCY,
            self.SERVER_CONFIG,
        )

        self._concurrency = self.CONCURRENCY
        self._loop = asyncio.get_event_loop()

        self._server_config = parse_server_config(self.SERVER_CONFIG)
        self._server = Server(self.NAME, self._server_config, loop=self._loop)
        self._semaphore = asyncio.locks.Semaphore(self.CONCURRENCY, loop=self._loop)

        # Running tasks cache
        self._running_tasks: Dict[TaskId, Any] = {}

        # Count the running tasks
        self._lock = asyncio.Lock()
        self._running_tasks_count = 0

        self._stop = False
        self._run_forever = run_forever

    @property
    def _cls_name(self) -> str:
        return self.__class__.__name__

    @property
    def stop(self) -> bool:
        return self._stop

    @stop.setter
    def stop(self, val):
        self._stop = val

    @property
    def concurrency(self) -> int:
        return self._concurrency

    @concurrency.setter
    def concurrency(self, concurrency: int):
        self._concurrency = concurrency
        self._semaphore = asyncio.locks.Semaphore(self.CONCURRENCY, loop=self._loop)

    def _create_task(self, task, task_type: TaskType = None) -> asyncio.Task:
        """Wrap `loop.create_task`
        We add an id to task if `task_id` is not None
        """

        task_name: Optional[str] = None
        if task_type:
            task_name = task_type.value

        return self._loop.create_task(task, name=task_name)

    def _cache_task(self, task_id: TaskId, task: Any) -> None:
        """Cache running task"""

        self._running_tasks[task_id] = task

    def _uncache_task(self, task_id: TaskId) -> None:
        """Remove running task from cache"""

        self._running_tasks.pop(task_id, None)

    async def make_task(self, task_id: TaskId, task: Any) -> None:
        """This function is the procedure of a task.

        The function must be implemented for handling every tasks
        """

        logger.error("%s: make_task is not implemented", self._cls_name)
        raise NotImplementedError("ChihuoLoop.make_task must be implemented")

    async def next_task(self) -> Optional[Tuple[TaskId, Any]]:
        """Get next task from backend"""

        async with self._lock:
            value = await self._server.pop_left()
            if not value:
                return None
            value = cast(Tuple[TaskId, Union[str, bytes]], value)

            self._running_tasks_count += 1

        task_id, raw_task = value
        task = stdjson.loads(raw_task)
        return task_id, task

    async def _task_loop(self) -> None:
        """Main Task loop

        The loop pops tasks from backend to execute asynchronously.
        """

        while True:
            await self._semaphore.acquire()

            # Stop the loop
            if self._stop:
                logger.info("%s: task_loop stop", self._cls_name)
                self._release()
                return

            item = await self.next_task()
            if item:
                task_id, task = item
                self._cache_task(task_id, task)
                self._create_task(
                    self._wrap_task(task_id, task), task_type=TaskType.Task
                )
            else:
                await self.sleep(1)
                # Release the task
                self._release()

    async def _wrap_task(self, task_id: TaskId, task: Any) -> None:
        """Wrap task
        Handle `asyncio.CancelledError`

        If the task fails, the task will be sent back to backend
        """

        try:
            await self.make_task(task_id, task)
        except BaseException as err:
            # Send task back to the backend
            await self.add_task(
                (task_id, task), ignore_running=False, direction=Direction.Reverse
            )
            if isinstance(err, asyncio.CancelledError):
                logger.warning("Cancelled: %s: task_id: %s", self._cls_name, task_id)
            else:
                logger.error(
                    "Task fails: %s: task_id: %s, task: %s, error: %s, traceback: %s",
                    self._cls_name,
                    task_id,
                    task,
                    err,
                    traceback.format_exc(),
                )
        finally:
            # Uncache the task
            self._uncache_task(task_id)
            # Release the semaphore for a completed task
            self._release()

        async with self._lock:
            self._running_tasks_count -= 1

            # The end event: running tasks is empty and the task queue is empty
            if not self._run_forever and self._running_tasks_count == 0:
                size = await self._server.size()
                if size == 0:
                    logger.info(
                        "`run_forever = %s`: It is the end. "
                        "There is no running tasks and the queue is empty.",
                        self._run_forever,
                    )
                    self._loop.stop()

    async def add_task(
        self,
        *pairs: Tuple[TaskId, Any],
        finished: bool = True,
        ignore_running: bool = False,
        direction: Direction = Direction.Forward
    ) -> None:
        """Add a task to backend server

        `pairs`: [(task_id1, task1), (task_id2, task2), ...]
            `pairs` are the information of tasks.

        `task_id`: str or bytes
            `task_id` is the unique identification for a task.
        `task`: object
            `task` is the information that can be dumped as json about the task.

        `finished`: bool
            if finished is true, we add the task only if the task is not finished.
            Else, we add the task whether or not the task is finished.
        `ignore_running`: bool
            if ignore_running is true, the task will be added to the task queue
            only if there is not a task which is running with same task_id
        `direction`: chihuo.Direction
            if direction is `Direction.Forward`, the task will be appended to the
            tail of queue, received at last. if direction is `Direction.Reverse`, the
            task will be appended to the head of queue, received at first.
        """

        if ignore_running:
            # Remove these tasks which is running
            pairs = tuple(pair for pair in pairs if pair[0] not in self._running_tasks)

        if finished:
            unfinished_pairs = []
            for task_id, task in pairs:
                v = await self._server.finished(task_id)
                if not v:
                    unfinished_pairs.append((task_id, task))
            pairs = tuple(unfinished_pairs)

        if not pairs:
            return

        # Serialize tasks
        pairs = tuple((task_id, serialize_json(task)) for task_id, task in pairs)

        if direction == Direction.Forward:
            await self._server.pushnx(*pairs)
        else:
            await self._server.pushnx_left(*pairs)

    async def task_finish(self, task_id: TaskId) -> None:
        """Set the task as finished

        If a task is set as finished, it CAN NOT be added to the backend at next time.
        """

        logger.info("%s: task_finish: %s", self._cls_name, task_id)

        await self._server.finish(task_id)

    async def task_unfinish(self, task_id: TaskId) -> None:
        """Set the task as unfinished

        If a task is set as unfinished, it CAN be added to the backend at next time.
        """

        logger.info("%s: task_unfinish: %s", self._cls_name, task_id)

        await self._server.unfinish(task_id)

    async def task_finished(self, task_id: TaskId) -> bool:
        """Check if the task is finished"""

        return await self._server.finished(task_id)

    async def task_exists(self, task_id: TaskId) -> bool:
        """Check if the task exists"""

        return await self._server.exists(task_id)

    def _release(self) -> None:
        """Release the semaphore"""

        self._semaphore.release()

    def length(self):
        """How many tasks are there"""

        return self._concurrency - self._semaphore._value

    async def sleep(self, second) -> None:
        await asyncio.sleep(second, loop=self._loop)

    def _run_(self):
        """The main function to run the factory"""

        logger.info("%s runs", self._cls_name)

        start = getattr(self, "start", None)
        if start and isinstance(start, types.MethodType):
            self._create_task(start(), task_type=TaskType.Start)

        self._create_task(self._task_loop(), task_type=TaskType.TaskLoop)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chihuo']

package_data = \
{'': ['*']}

install_requires = \
['lodis-py>=0.2.2,<0.3.0']

setup_kwargs = {
    'name': 'chihuo',
    'version': '0.3.4',
    'description': 'chihuo: An Asynchronous Task Running Engine',
    'long_description': '# chihuo: An Asynchronous Task Running Engine\n\nchihuo is an asynchronous task running engine. It runs asynchronous tasks concurrently\nfrom a queue which can be at memory, Redis, Lodis or Some MQ.\n\n## The Loop\n\n`ChihuoLoop` is the asynchronous engine core which responses to connect the backend queue\nand distributes every tasks to the global event loop.\n\nFor a process, the chihuo create only one global event loop to run all asynchronous\ntasks and the loop is running forever.\n\n- `ChihuoLoop.start` method\n\n  When a `ChihuoLoop` starts, it needs to add some primary tasks to the backend queue.\n  The `start` function is the place to fire these priorities.\n\n- `ChihuoLoop.add_task` method\n\n  At anywhere and anytime, we can use the `add_task` method to add a task to backend queue.\n  The task discription must be the tuple as `(task_id: str, task_info: jsonifiable object_)`\n  e.g. `add_task(("task1", [1, 1+10]), ("task2", [2, 2+10]))`\n\n- `ChihuoLoop.make_task` method\n\n  Each classes which have implemented the `ChihuoLoop` class must to implement the `make_task`\n  method for receiving the task\'s information that is added by `ChihuoLoop.add_task` to\n  the backend queue.\n\n- `ChihuoLoop.task_finish` method\n\n  When a task is finished and is unneeded at future, we can tag the task as the `finished` state\n  using the `task_finish` method. e.g. `task_finish(task_id)`. The task that is at `finished`\n  state can\'t be add to backend queue again.\n\n- `ChihuoLoop.task_unfinish` method\n\n  When a task is at the `finished` state and we need to add it to queue again, we can use\n  the `task_unfinish` method to set the task state to `unfinish`. Then, we can use\n  `ChihuoLoop.add_task` to add the task to queue again.\n\n- `ChihuoLoop.stop` property\n\n  When the property `stop` to be set as True, the `ChihuoLoop` will stop pull the tasks from\n  queue and exit.\n\n## Demo\n\n```python\n# filename: abc.py\n\nfrom chihuo import ChihuoLoop\nfrom chihuo.common import SERVER_DEFAULT_CONFIG_PATH\nimport chihuo\n\n\nclass MyTasksA(ChihuoLoop):\n\n    NAME = "my-tasksA"  # The unique id for backend queue server\n    CONCURRENCY = 10   # The number of tasks running concurrently\n    SERVER_CONFIG = SERVER_DEFAULT_CONFIG_PATH\n\n    def __init__(self):\n        super().__init__()\n\n    async def start(self):\n        for id in range(10):\n            await self.add_task((str(i), {i ** 2}))\n\n    async def make_task(self, task_id, task):\n        print(f"my-task-A: {task_id}")\n        print(f"power of {i} is {task}")\n\n        # Talk chihuo that the task is finished\n        await self.task_finish(i)\n\n\nclass MyTasksB(ChihuoLoop):\n\n    NAME = "my-tasksB"  # The unique id for the loop\n    CONCURRENCY = 10   # The number of tasks running concurrently\n    SERVER_CONFIG = SERVER_DEFAULT_CONFIG_PATH\n\n    def __init__(self):\n        super().__init__()\n\n    async def start(self):\n        for id in range(10):\n            await self.add_task((str(i), {i ** 2}))\n\n    async def make_task(self, task_id, task):\n        print(f"my-task-B: {task_id}")\n        print(f"power of {i} is {task}")\n\n        # Talk chihuo that the task is finished\n        await self.task_finish(i)\n\n\nif __name__ == \'__main__\':\n    chihuo.run(MyTasksA, MyTasksB)\n```\n\nTo Run the project, firstly we need to launch a backend queue server.\nHere, we use a Lodis instance as default server.\n\nStart a Lodis server:\n\n```\nLODIS_DB_PATH=/data/test-lodis LODIS_IP_PORT="127.0.0.1:8311" nohup /path/to/lodis &\n```\n\nThen, we write a configure file as `./server.json` (the `SERVER_DEFAULT_CONFIG_PATH`)\nto connect the backend for our `MyTasksA` and `MyTasksB`.\n\n```\nbackend=lodis\nip=127.0.0.1\nport=8311\n```\n\nNow, we can run the script.\n\n```\npython3 abc.py\n```\n\n## Shutdown the Tasks\n\nDefaultly, the global loop is running forever and we need to shutdown the loop by sending\na `INT (interrupt)` signal to the process. After the process received The `INT` signal, it will\nset the property `stop` of instance which is implemented from `ChihuoLoop` to True and the\ninstance will stop all running tasks and send them back to queue, finally exit.\n',
    'author': 'PeterDing',
    'author_email': 'dfhayst@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PeterDing/chihuo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
